@@ -10,23 +10,19 @@
 
 #include "access/gist.h"
 #include "access/rtree.h"
-#include "utils/elog.h"
-#include "utils/palloc.h"
 #include "utils/builtins.h"
 
 #include "segdata.h"
 
-#define max(a,b)		((a) >	(b) ? (a) : (b))
-#define min(a,b)		((a) <= (b) ? (a) : (b))
-#define abs(a)			((a) <	(0) ? (-a) : (a))
-
 /*
 #define GIST_DEBUG
 #define GIST_QUERY_DEBUG
 */
 
-extern void set_parse_buffer(char *str);
 extern int	seg_yyparse();
+extern void seg_yyerror(const char *message);
+extern void seg_scanner_init(const char *str);
+extern void seg_scanner_finish(void);
 
 /*
 extern int	 seg_yydebug;
@@ -48,10 +44,10 @@ bool		gseg_consistent(GISTENTRY *entry, SEG * query, StrategyNumber strategy);
 GISTENTRY  *gseg_compress(GISTENTRY *entry);
 GISTENTRY  *gseg_decompress(GISTENTRY *entry);
 float	   *gseg_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result);
-GIST_SPLITVEC *gseg_picksplit(bytea *entryvec, GIST_SPLITVEC *v);
+GIST_SPLITVEC *gseg_picksplit(GistEntryVector *entryvec, GIST_SPLITVEC *v);
 bool		gseg_leaf_consistent(SEG * key, SEG * query, StrategyNumber strategy);
 bool		gseg_internal_consistent(SEG * key, SEG * query, StrategyNumber strategy);
-SEG		   *gseg_union(bytea *entryvec, int *sizep);
+SEG		   *gseg_union(GistEntryVector *entryvec, int *sizep);
 SEG		   *gseg_binary_union(SEG * r1, SEG * r2, int *sizep);
 bool	   *gseg_same(SEG * b1, SEG * b2, bool *result);
 
@@ -101,27 +97,16 @@ seg_in(char *str)
 {
 	SEG		   *result = palloc(sizeof(SEG));
 
-	set_parse_buffer(str);
+	seg_scanner_init(str);
 
-	/*
-	 * seg_yydebug = 1;
-	 */
 	if (seg_yyparse(result) != 0)
-	{
-		pfree(result);
-		return NULL;
-	}
+		seg_yyerror("bogus input");
+
+	seg_scanner_finish();
+
 	return (result);
 }
 
-/*
- * You might have noticed a slight inconsistency between the following
- * declaration and the SQL definition:
- *	   CREATE FUNCTION seg_out(opaque) RETURNS opaque ...
- * The reason is that the argument passed into seg_out is really just a
- * pointer. POSTGRES thinks all output functions are:
- *	   char *out_func(char *);
- */
 char *
 seg_out(SEG * seg)
 {
@@ -139,8 +124,7 @@ seg_out(SEG * seg)
 	if (seg->lower == seg->upper && seg->l_ext == seg->u_ext)
 	{
 		/*
-		 * indicates that this interval was built by seg_in off a single
-		 * point
+		 * indicates that this interval was built by seg_in off a single point
 		 */
 		p += restore(p, seg->lower, seg->l_sigd);
 	}
@@ -219,7 +203,7 @@ gseg_consistent(GISTENTRY *entry,
 				StrategyNumber strategy)
 {
 	/*
-	 * * if entry is not leaf, use gseg_internal_consistent, * else use
+	 * if entry is not leaf, use gseg_internal_consistent, else use
 	 * gseg_leaf_consistent
 	 */
 	if (GIST_LEAF(entry))
@@ -233,7 +217,7 @@ gseg_consistent(GISTENTRY *entry,
 ** returns the minimal bounding seg that encloses all the entries in entryvec
 */
 SEG *
-gseg_union(bytea *entryvec, int *sizep)
+gseg_union(GistEntryVector *entryvec, int *sizep)
 {
 	int			numranges,
 				i;
@@ -244,17 +228,15 @@ gseg_union(bytea *entryvec, int *sizep)
 	fprintf(stderr, "union\n");
 #endif
 
-	numranges = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
-	tmp = (SEG *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[0].key);
+	numranges = entryvec->n;
+	tmp = (SEG *) DatumGetPointer(entryvec->vector[0].key);
 	*sizep = sizeof(SEG);
 
 	for (i = 1; i < numranges; i++)
 	{
 		out = gseg_binary_union(tmp, (SEG *)
-			   DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[i].key),
+								DatumGetPointer(entryvec->vector[i].key),
 								sizep);
-		if (i > 1)
-			pfree(tmp);
 		tmp = out;
 	}
 
@@ -293,7 +275,6 @@ gseg_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result)
 	rt_seg_size(ud, &tmp1);
 	rt_seg_size((SEG *) DatumGetPointer(origentry->key), &tmp2);
 	*result = tmp1 - tmp2;
-	pfree(ud);
 
 #ifdef GIST_DEBUG
 	fprintf(stderr, "penalty\n");
@@ -310,7 +291,7 @@ gseg_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result)
 ** We use Guttman's poly time split algorithm
 */
 GIST_SPLITVEC *
-gseg_picksplit(bytea *entryvec,
+gseg_picksplit(GistEntryVector *entryvec,
 			   GIST_SPLITVEC *v)
 {
 	OffsetNumber i,
@@ -343,7 +324,7 @@ gseg_picksplit(bytea *entryvec,
 	fprintf(stderr, "picksplit\n");
 #endif
 
-	maxoff = ((VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY)) - 2;
+	maxoff = entryvec->n - 2;
 	nbytes = (maxoff + 2) * sizeof(OffsetNumber);
 	v->spl_left = (OffsetNumber *) palloc(nbytes);
 	v->spl_right = (OffsetNumber *) palloc(nbytes);
@@ -353,10 +334,10 @@ gseg_picksplit(bytea *entryvec,
 
 	for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i))
 	{
-		datum_alpha = (SEG *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[i].key);
+		datum_alpha = (SEG *) DatumGetPointer(entryvec->vector[i].key);
 		for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j))
 		{
-			datum_beta = (SEG *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[j].key);
+			datum_beta = (SEG *) DatumGetPointer(entryvec->vector[j].key);
 
 			/* compute the wasted space by unioning these guys */
 			/* size_waste = size_union - size_inter; */
@@ -366,16 +347,9 @@ gseg_picksplit(bytea *entryvec,
 			rt_seg_size(inter_d, &size_inter);
 			size_waste = size_union - size_inter;
 
-			pfree(union_d);
-
-			if (inter_d != (SEG *) NULL)
-				pfree(inter_d);
-
 			/*
-			 * are these a more promising split that what we've already
-			 * seen?
+			 * are these a more promising split that what we've already seen?
 			 */
-
 			if (size_waste > waste || firsttime)
 			{
 				waste = size_waste;
@@ -391,32 +365,32 @@ gseg_picksplit(bytea *entryvec,
 	right = v->spl_right;
 	v->spl_nright = 0;
 
-	datum_alpha = (SEG *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[seed_1].key);
+	datum_alpha = (SEG *) DatumGetPointer(entryvec->vector[seed_1].key);
 	datum_l = seg_union(datum_alpha, datum_alpha);
 	rt_seg_size(datum_l, &size_l);
-	datum_beta = (SEG *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[seed_2].key);
+	datum_beta = (SEG *) DatumGetPointer(entryvec->vector[seed_2].key);
 	datum_r = seg_union(datum_beta, datum_beta);
 	rt_seg_size(datum_r, &size_r);
 
 	/*
-	 * Now split up the regions between the two seeds.	An important
-	 * property of this split algorithm is that the split vector v has the
-	 * indices of items to be split in order in its left and right
-	 * vectors.  We exploit this property by doing a merge in the code
-	 * that actually splits the page.
+	 * Now split up the regions between the two seeds.	An important property
+	 * of this split algorithm is that the split vector v has the indices of
+	 * items to be split in order in its left and right vectors.  We exploit
+	 * this property by doing a merge in the code that actually splits the
+	 * page.
 	 *
-	 * For efficiency, we also place the new index tuple in this loop. This
-	 * is handled at the very end, when we have placed all the existing
-	 * tuples and i == maxoff + 1.
+	 * For efficiency, we also place the new index tuple in this loop. This is
+	 * handled at the very end, when we have placed all the existing tuples
+	 * and i == maxoff + 1.
 	 */
 
 	maxoff = OffsetNumberNext(maxoff);
 	for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i))
 	{
 		/*
-		 * If we've already decided where to place this item, just put it
-		 * on the right list.  Otherwise, we need to figure out which page
-		 * needs the least enlargement in order to store the item.
+		 * If we've already decided where to place this item, just put it on
+		 * the right list.	Otherwise, we need to figure out which page needs
+		 * the least enlargement in order to store the item.
 		 */
 
 		if (i == seed_1)
@@ -433,7 +407,7 @@ gseg_picksplit(bytea *entryvec,
 		}
 
 		/* okay, which page needs least enlargement? */
-		datum_alpha = (SEG *) DatumGetPointer(((GISTENTRY *) VARDATA(entryvec))[i].key);
+		datum_alpha = (SEG *) DatumGetPointer(entryvec->vector[i].key);
 		union_dl = seg_union(datum_l, datum_alpha);
 		union_dr = seg_union(datum_r, datum_alpha);
 		rt_seg_size(union_dl, &size_alpha);
@@ -442,8 +416,6 @@ gseg_picksplit(bytea *entryvec,
 		/* pick which page to add it to */
 		if (size_alpha - size_l < size_beta - size_r)
 		{
-			pfree(datum_l);
-			pfree(union_dr);
 			datum_l = union_dl;
 			size_l = size_alpha;
 			*left++ = i;
@@ -451,8 +423,6 @@ gseg_picksplit(bytea *entryvec,
 		}
 		else
 		{
-			pfree(datum_r);
-			pfree(union_dl);
 			datum_r = union_dr;
 			size_r = size_alpha;
 			*right++ = i;
@@ -545,15 +515,19 @@ gseg_internal_consistent(SEG * key,
 	switch (strategy)
 	{
 		case RTLeftStrategyNumber:
+			retval = (bool) !seg_over_right(key, query);
+			break;
 		case RTOverLeftStrategyNumber:
-			retval = (bool) seg_over_left(key, query);
+			retval = (bool) !seg_right(key, query);
 			break;
 		case RTOverlapStrategyNumber:
 			retval = (bool) seg_overlap(key, query);
 			break;
 		case RTOverRightStrategyNumber:
+			retval = (bool) !seg_left(key, query);
+			break;
 		case RTRightStrategyNumber:
-			retval = (bool) seg_right(key, query);
+			retval = (bool) !seg_over_left(key, query);
 			break;
 		case RTSameStrategyNumber:
 		case RTContainsStrategyNumber:
@@ -614,12 +588,12 @@ seg_overlap(SEG * a, SEG * b)
 		);
 }
 
-/*	seg_overleft -- is the right edge of (a) located to the left of the right edge of (b)?
+/*	seg_overleft -- is the right edge of (a) located at or left of the right edge of (b)?
  */
 bool
 seg_over_left(SEG * a, SEG * b)
 {
-	return (a->upper <= b->upper && !seg_left(a, b) && !seg_right(a, b));
+	return (a->upper <= b->upper);
 }
 
 /*	seg_left -- is (a) entirely on the left of (b)?
@@ -638,12 +612,12 @@ seg_right(SEG * a, SEG * b)
 	return (a->lower > b->upper);
 }
 
-/*	seg_overright -- is the left edge of (a) located to the right of the left edge of (b)?
+/*	seg_overright -- is the left edge of (a) located at or right of the left edge of (b)?
  */
 bool
 seg_over_right(SEG * a, SEG * b)
 {
-	return (a->lower >= b->lower && !seg_left(a, b) && !seg_right(a, b));
+	return (a->lower >= b->lower);
 }
 
 
@@ -730,7 +704,7 @@ rt_seg_size(SEG * a, float *size)
 	if (a == (SEG *) NULL || a->upper <= a->lower)
 		*size = 0.0;
 	else
-		*size = (float) abs(a->upper - a->lower);
+		*size = (float) Abs(a->upper - a->lower);
 
 	return;
 }
@@ -742,7 +716,7 @@ seg_size(SEG * a)
 
 	result = (float *) palloc(sizeof(float));
 
-	*result = (float) abs(a->upper - a->lower);
+	*result = (float) Abs(a->upper - a->lower);
 
 	return (result);
 }
@@ -766,7 +740,7 @@ seg_cmp(SEG * a, SEG * b)
 	 * a->lower == b->lower, so consider type of boundary.
 	 *
 	 * A '-' lower bound is < any other kind (this could only be relevant if
-	 * -HUGE is used as a regular data value). A '<' lower bound is < any
+	 * -HUGE_VAL is used as a regular data value). A '<' lower bound is < any
 	 * other kind except '-'. A '>' lower bound is > any other kind.
 	 */
 	if (a->l_ext != b->l_ext)
@@ -788,8 +762,7 @@ seg_cmp(SEG * a, SEG * b)
 	/*
 	 * For other boundary types, consider # of significant digits first.
 	 */
-	if (a->l_sigd < b->l_sigd)	/* (a) is blurred and is likely to include
-								 * (b) */
+	if (a->l_sigd < b->l_sigd)	/* (a) is blurred and is likely to include (b) */
 		return -1;
 	if (a->l_sigd > b->l_sigd)	/* (a) is less blurred and is likely to be
 								 * included in (b) */
@@ -806,7 +779,7 @@ seg_cmp(SEG * a, SEG * b)
 		if (b->l_ext == '~')
 			return 1;
 		/* can't get here unless data is corrupt */
-		elog(ERROR, "seg_cmp: bogus lower boundary types %d %d",
+		elog(ERROR, "bogus lower boundary types %d %d",
 			 (int) a->l_ext, (int) b->l_ext);
 	}
 
@@ -824,7 +797,7 @@ seg_cmp(SEG * a, SEG * b)
 	 * a->upper == b->upper, so consider type of boundary.
 	 *
 	 * A '-' upper bound is > any other kind (this could only be relevant if
-	 * HUGE is used as a regular data value). A '<' upper bound is < any
+	 * HUGE_VAL is used as a regular data value). A '<' upper bound is < any
 	 * other kind. A '>' upper bound is > any other kind except '-'.
 	 */
 	if (a->u_ext != b->u_ext)
@@ -844,11 +817,10 @@ seg_cmp(SEG * a, SEG * b)
 	}
 
 	/*
-	 * For other boundary types, consider # of significant digits first.
-	 * Note result here is converse of the lower-boundary case.
+	 * For other boundary types, consider # of significant digits first. Note
+	 * result here is converse of the lower-boundary case.
 	 */
-	if (a->u_sigd < b->u_sigd)	/* (a) is blurred and is likely to include
-								 * (b) */
+	if (a->u_sigd < b->u_sigd)	/* (a) is blurred and is likely to include (b) */
 		return 1;
 	if (a->u_sigd > b->u_sigd)	/* (a) is less blurred and is likely to be
 								 * included in (b) */
@@ -865,7 +837,7 @@ seg_cmp(SEG * a, SEG * b)
 		if (b->u_ext == '~')
 			return -1;
 		/* can't get here unless data is corrupt */
-		elog(ERROR, "seg_cmp: bogus upper boundary types %d %d",
+		elog(ERROR, "bogus upper boundary types %d %d",
 			 (int) a->u_ext, (int) b->u_ext);
 	}
 
@@ -890,7 +862,6 @@ seg_gt(SEG * a, SEG * b)
 	return seg_cmp(a, b) > 0;
 }
 
-
 bool
 seg_ge(SEG * a, SEG * b)
 {
@@ -933,17 +904,17 @@ restore(char *result, float val, int n)
 				sign;
 
 	/*
-	 * put a cap on the number of siugnificant digits to avoid nonsense in
-	 * the output
+	 * put a cap on the number of siugnificant digits to avoid nonsense in the
+	 * output
 	 */
-	n = min(n, FLT_DIG);
+	n = Min(n, FLT_DIG);
 
 	/* remember the sign */
 	sign = (val < 0 ? 1 : 0);
 
-	efmt[5] = '0' + (n - 1) % 10;		/* makes %-15.(n-1)e -- this
-										 * format guarantees that the
-										 * exponent is always present */
+	efmt[5] = '0' + (n - 1) % 10;		/* makes %-15.(n-1)e -- this format
+										 * guarantees that the exponent is
+										 * always present */
 
 	sprintf(result, efmt, val);
 
@@ -958,15 +929,15 @@ restore(char *result, float val, int n)
 	if (exp == 0)
 	{
 		/* use the supplied mantyssa with sign */
-		strcpy((char *) index(result, 'e'), "");
+		strcpy((char *) strchr(result, 'e'), "");
 	}
 	else
 	{
-		if (abs(exp) <= 4)
+		if (Abs(exp) <= 4)
 		{
 			/*
-			 * remove the decimal point from the mantyssa and write the
-			 * digits to the buf array
+			 * remove the decimal point from the mantyssa and write the digits
+			 * to the buf array
 			 */
 			for (p = result + sign, i = 10, dp = 0; *p != 'e'; p++, i++)
 			{
@@ -985,10 +956,9 @@ restore(char *result, float val, int n)
 				if (dp - 10 + exp >= n)
 				{
 					/*
-					 * the decimal point is behind the last significant
-					 * digit; the digits in between must be converted to
-					 * the exponent and the decimal point placed after the
-					 * first digit
+					 * the decimal point is behind the last significant digit;
+					 * the digits in between must be converted to the exponent
+					 * and the decimal point placed after the first digit
 					 */
 					exp = dp - 10 + exp - n;
 					buf[10 + n] = '\0';
@@ -1003,8 +973,8 @@ restore(char *result, float val, int n)
 					}
 
 					/*
-					 * adjust the exponent by the number of digits after
-					 * the decimal point
+					 * adjust the exponent by the number of digits after the
+					 * decimal point
 					 */
 					if (n > 1)
 						sprintf(&buf[11 + n], "e%d", exp + n - 1);
@@ -1050,7 +1020,7 @@ restore(char *result, float val, int n)
 			}
 		}
 
-		/* do nothing for abs(exp) > 4; %e must be OK */
+		/* do nothing for Abs(exp) > 4; %e must be OK */
 		/* just get rid of zeroes after [eE]- and +zeroes after [Ee]. */
 
 		/* ... this is not done yet. */
