@@ -55,6 +55,7 @@ Datum		to_tsquery_current(PG_FUNCTION_ARGS);
 /* parser's states */
 #define WAITOPERAND 1
 #define WAITOPERATOR	2
+#define WAITFIRSTOPERAND 3
 
 /*
  * node of query tree, also used
@@ -137,6 +138,7 @@ gettoken_query(QPRS_STATE * state, int4 *val, int4 *lenval, char **strval, int2
 	{
 		switch (state->state)
 		{
+			case WAITFIRSTOPERAND:
 			case WAITOPERAND:
 				if (*(state->buf) == '!')
 				{
@@ -161,14 +163,16 @@ gettoken_query(QPRS_STATE * state, int4 *val, int4 *lenval, char **strval, int2
 				else if (*(state->buf) != ' ')
 				{
 					state->valstate.prsbuf = state->buf;
-					state->state = WAITOPERATOR;
 					if (gettoken_tsvector(&(state->valstate)))
 					{
 						*strval = state->valstate.word;
 						*lenval = state->valstate.curpos - state->valstate.word;
 						state->buf = get_weight(state->valstate.prsbuf, weight);
+						state->state = WAITOPERATOR;
 						return VAL;
 					}
+					else if (state->state == WAITFIRSTOPERAND)
+						return END;
 					else
 						ereport(ERROR,
 								(errcode(ERRCODE_SYNTAX_ERROR),
@@ -240,7 +244,7 @@ pushval_asis(QPRS_STATE * state, int type, char *strval, int lenval, int2 weight
 				(errcode(ERRCODE_SYNTAX_ERROR),
 				 errmsg("word is too long")));
 
-	pushquery(state, type, crc32_sz((uint8 *) strval, lenval),
+	pushquery(state, type, crc32_sz(strval, lenval),
 			  state->curop - state->op, lenval, weight);
 
 	while (state->curop - state->op + lenval + 1 >= state->lenop)
@@ -267,6 +271,11 @@ pushval_morph(QPRS_STATE * state, int typeval, char *strval, int lenval, int2 we
 {
 	int4		count = 0;
 	PRSTEXT		prs;
+	uint32		variant,
+				pos,
+				cntvar = 0,
+				cntpos = 0,
+				cnt = 0;
 
 	prs.lenwords = 32;
 	prs.curwords = 0;
@@ -275,17 +284,44 @@ pushval_morph(QPRS_STATE * state, int typeval, char *strval, int lenval, int2 we
 
 	parsetext_v2(findcfg(state->cfg_id), &prs, strval, lenval);
 
-	for (count = 0; count < prs.curwords; count++)
+	if (prs.curwords > 0)
 	{
-		pushval_asis(state, VAL, prs.words[count].word, prs.words[count].len, weight);
-		pfree(prs.words[count].word);
-		if (count)
-			pushquery(state, OPR, (int4) '&', 0, 0, 0);
-	}
-	pfree(prs.words);
 
-	/* XXX */
-	if (prs.curwords == 0)
+		while (count < prs.curwords)
+		{
+			pos = prs.words[count].pos.pos;
+			cntvar = 0;
+			while (count < prs.curwords && pos == prs.words[count].pos.pos)
+			{
+				variant = prs.words[count].nvariant;
+
+				cnt = 0;
+				while (count < prs.curwords && pos == prs.words[count].pos.pos && variant == prs.words[count].nvariant)
+				{
+
+					pushval_asis(state, VAL, prs.words[count].word, prs.words[count].len, weight);
+					pfree(prs.words[count].word);
+					if (cnt)
+						pushquery(state, OPR, (int4) '&', 0, 0, 0);
+					cnt++;
+					count++;
+				}
+
+				if (cntvar)
+					pushquery(state, OPR, (int4) '|', 0, 0, 0);
+				cntvar++;
+			}
+
+			if (cntpos)
+				pushquery(state, OPR, (int4) '&', 0, 0, 0);
+
+			cntpos++;
+		}
+
+		pfree(prs.words);
+
+	}
+	else
 		pushval_asis(state, VALSTOP, NULL, 0, 0);
 }
 
@@ -296,13 +332,13 @@ pushval_morph(QPRS_STATE * state, int typeval, char *strval, int lenval, int2 we
 static int4
 makepol(QPRS_STATE * state, void (*pushval) (QPRS_STATE *, int, char *, int, int2))
 {
-	int4		val,
+	int4		val = 0,
 				type;
-	int4		lenval;
-	char	   *strval;
+	int4		lenval = 0;
+	char	   *strval = NULL;
 	int4		stack[STACKDEPTH];
 	int4		lenstack = 0;
-	int2		weight;
+	int2		weight = 0;
 
 	while ((type = gettoken_query(state, &val, &lenval, &strval, &weight)) != END)
 	{
@@ -398,7 +434,7 @@ checkclass_str(CHKVAL * chkval, WordEntry * val, ITEM * item)
 
 	while (len--)
 	{
-		if (item->weight & (1 << ptr->weight))
+		if (item->weight & (1 << WEP_GETWEIGHT(*ptr)))
 			return true;
 		ptr++;
 	}
@@ -566,7 +602,7 @@ static QUERYTYPE *
 
 	/* init state */
 	state.buf = buf;
-	state.state = WAITOPERAND;
+	state.state = WAITFIRSTOPERAND;
 	state.count = 0;
 	state.num = 0;
 	state.str = NULL;
@@ -587,9 +623,13 @@ static QUERYTYPE *
 	makepol(&state, pushval);
 	pfree(state.valstate.word);
 	if (!state.num)
-		ereport(ERROR,
-				(errcode(ERRCODE_SYNTAX_ERROR),
-				 errmsg("empty query")));
+	{
+		elog(NOTICE, "Query doesn't contain lexem(s)");
+		query = (QUERYTYPE *) palloc(HDRSIZEQT);
+		query->len = HDRSIZEQT;
+		query->size = 0;
+		return query;
+	}
 
 	/* make finish struct */
 	commonlen = COMPUTESIZE(state.num, state.sumlen);
@@ -659,12 +699,12 @@ typedef struct
 }	INFIX;
 
 #define RESIZEBUF(inf,addsize) \
-while( ( inf->cur - inf->buf ) + addsize + 1 >= inf->buflen ) \
+while( ( (inf)->cur - (inf)->buf ) + (addsize) + 1 >= (inf)->buflen ) \
 { \
-	int4 len = inf->cur - inf->buf; \
-	inf->buflen *= 2; \
-	inf->buf = (char*) repalloc( (void*)inf->buf, inf->buflen ); \
-	inf->cur = inf->buf + len; \
+	int4 len = (inf)->cur - (inf)->buf; \
+	(inf)->buflen *= 2; \
+	(inf)->buf = (char*) repalloc( (void*)(inf)->buf, (inf)->buflen ); \
+	(inf)->cur = (inf)->buf + len; \
 }
 
 /*
@@ -875,6 +915,10 @@ to_tsquery(PG_FUNCTION_ARGS)
 	PG_FREE_IF_COPY(in, 1);
 
 	query = queryin(str, pushval_morph, PG_GETARG_INT32(0));
+
+	if (query->size == 0)
+		PG_RETURN_POINTER(query);
+
 	res = clean_fakeval_v2(GETQUERY(query), &len);
 	if (!res)
 	{
