@@ -25,7 +25,7 @@
  * scenario of this sort is "out of memory"; and it's also the nastiest
  * to handle because we'd likely also run out of memory while trying to
  * report this error!  Our escape hatch for this case is to reset the
- * ErrorContext to empty before trying to process the inner error.  Since
+ * ErrorContext to empty before trying to process the inner error.	Since
  * ErrorContext is guaranteed to have at least 8K of space in it (see mcxt.c),
  * we should be able to process an "out of memory" message successfully.
  * Since we lose the prior error state due to the reset, we won't be able
@@ -42,7 +42,7 @@
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/utils/error/elog.c,v 1.155.4.7 2007/07/21 22:12:24 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/error/elog.c,v 1.167.2.2 2007/02/11 15:12:48 mha Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -75,7 +75,7 @@ ErrorContextCallback *error_context_stack = NULL;
 
 sigjmp_buf *PG_exception_stack = NULL;
 
-extern bool redirection_done;
+extern pid_t SysLoggerPID;
 
 /* GUC parameters */
 PGErrorVerbosity Log_error_verbosity = PGERROR_VERBOSE;
@@ -83,8 +83,9 @@ char	   *Log_line_prefix = NULL;		/* format for extra log line info */
 int			Log_destination = LOG_DESTINATION_STDERR;
 
 #ifdef HAVE_SYSLOG
-char	   *Syslog_facility;	/* openlog() parameters */
-char	   *Syslog_ident;
+static bool openlog_done = false;
+static char *syslog_ident = NULL;
+static int	syslog_facility = LOG_LOCAL0;
 
 static void write_syslog(int level, const char *line);
 #endif
@@ -121,7 +122,7 @@ static char *expand_fmt_string(const char *fmt, ErrorData *edata);
 static const char *useful_strerror(int errnum);
 static const char *error_severity(int elevel);
 static void append_with_tabs(StringInfo buf, const char *str);
-static void write_pipe_chunks(int fd, char *data, int len);
+
 
 /*
  * errstart --- begin an error-reporting cycle
@@ -190,8 +191,8 @@ errstart(int elevel, const char *filename, int lineno,
 
 	/*
 	 * Now decide whether we need to process this report at all; if it's
-	 * warning or less and not enabled for logging, just return FALSE
-	 * without starting up any error logging machinery.
+	 * warning or less and not enabled for logging, just return FALSE without
+	 * starting up any error logging machinery.
 	 */
 
 	/* Determine whether message is enabled for server log output */
@@ -225,7 +226,7 @@ errstart(int elevel, const char *filename, int lineno,
 	}
 
 	/* Determine whether message is enabled for client output */
-	if (whereToSendOutput == Remote && elevel != COMMERROR)
+	if (whereToSendOutput == DestRemote && elevel != COMMERROR)
 	{
 		/*
 		 * client_min_messages is honored only after we complete the
@@ -259,15 +260,10 @@ errstart(int elevel, const char *filename, int lineno,
 
 		/*
 		 * If we recurse more than once, the problem might be something broken
-		 * in a context traceback routine.  Abandon them too.  We also
-		 * abandon attempting to print the error statement (which, if long,
-		 * could itself be the source of the recursive failure).
+		 * in a context traceback routine.	Abandon them too.
 		 */
 		if (recursion_depth > 2)
-		{
 			error_context_stack = NULL;
-			debug_query_string = NULL;
-		}
 	}
 	if (++errordata_stack_depth >= ERRORDATA_STACK_SIZE)
 	{
@@ -323,15 +319,15 @@ errfinish(int dummy,...)
 	CHECK_STACK_DEPTH();
 
 	/*
-	 * Do processing in ErrorContext, which we hope has enough reserved
-	 * space to report an error.
+	 * Do processing in ErrorContext, which we hope has enough reserved space
+	 * to report an error.
 	 */
 	oldcontext = MemoryContextSwitchTo(ErrorContext);
 
 	/*
 	 * Call any context callback functions.  Errors occurring in callback
-	 * functions will be treated as recursive errors --- this ensures we
-	 * will avoid infinite recursion (see errstart).
+	 * functions will be treated as recursive errors --- this ensures we will
+	 * avoid infinite recursion (see errstart).
 	 */
 	for (econtext = error_context_stack;
 		 econtext != NULL;
@@ -340,34 +336,32 @@ errfinish(int dummy,...)
 
 	/*
 	 * If ERROR (not more nor less) we pass it off to the current handler.
-	 * Printing it and popping the stack is the responsibility of
-	 * the handler.
+	 * Printing it and popping the stack is the responsibility of the handler.
 	 */
 	if (elevel == ERROR)
 	{
 		/*
-		 * We do some minimal cleanup before longjmp'ing so that handlers
-		 * can execute in a reasonably sane state.
+		 * We do some minimal cleanup before longjmp'ing so that handlers can
+		 * execute in a reasonably sane state.
 		 */
 
 		/* This is just in case the error came while waiting for input */
 		ImmediateInterruptOK = false;
 
 		/*
-		 * Reset InterruptHoldoffCount in case we ereport'd from
-		 * inside an interrupt holdoff section.  (We assume here that
-		 * no handler will itself be inside a holdoff section.	If
-		 * necessary, such a handler could save and restore
-		 * InterruptHoldoffCount for itself, but this should make life
-		 * easier for most.)
+		 * Reset InterruptHoldoffCount in case we ereport'd from inside an
+		 * interrupt holdoff section.  (We assume here that no handler will
+		 * itself be inside a holdoff section.	If necessary, such a handler
+		 * could save and restore InterruptHoldoffCount for itself, but this
+		 * should make life easier for most.)
 		 */
 		InterruptHoldoffCount = 0;
 
-		CritSectionCount = 0;		/* should be unnecessary, but... */
+		CritSectionCount = 0;	/* should be unnecessary, but... */
 
 		/*
-		 * Note that we leave CurrentMemoryContext set to ErrorContext.
-		 * The handler should reset it to something else soon.
+		 * Note that we leave CurrentMemoryContext set to ErrorContext. The
+		 * handler should reset it to something else soon.
 		 */
 
 		recursion_depth--;
@@ -377,14 +371,13 @@ errfinish(int dummy,...)
 	/*
 	 * If we are doing FATAL or PANIC, abort any old-style COPY OUT in
 	 * progress, so that we can report the message before dying.  (Without
-	 * this, pq_putmessage will refuse to send the message at all, which
-	 * is what we want for NOTICE messages, but not for fatal exits.) This
-	 * hack is necessary because of poor design of old-style copy
-	 * protocol.  Note we must do this even if client is fool enough to
-	 * have set client_min_messages above FATAL, so don't look at
-	 * output_to_client.
+	 * this, pq_putmessage will refuse to send the message at all, which is
+	 * what we want for NOTICE messages, but not for fatal exits.) This hack
+	 * is necessary because of poor design of old-style copy protocol.	Note
+	 * we must do this even if client is fool enough to have set
+	 * client_min_messages above FATAL, so don't look at output_to_client.
 	 */
-	if (elevel >= FATAL && whereToSendOutput == Remote)
+	if (elevel >= FATAL && whereToSendOutput == DestRemote)
 		pq_endcopyout(true);
 
 	/* Emit the message to the right places */
@@ -419,28 +412,27 @@ errfinish(int dummy,...)
 		ImmediateInterruptOK = false;
 
 		/*
-		 * If we just reported a startup failure, the client will
-		 * disconnect on receiving it, so don't send any more to the
-		 * client.
+		 * If we just reported a startup failure, the client will disconnect
+		 * on receiving it, so don't send any more to the client.
 		 */
-		if (PG_exception_stack == NULL && whereToSendOutput == Remote)
-			whereToSendOutput = None;
+		if (PG_exception_stack == NULL && whereToSendOutput == DestRemote)
+			whereToSendOutput = DestNone;
 
 		/*
 		 * fflush here is just to improve the odds that we get to see the
-		 * error message, in case things are so hosed that proc_exit
-		 * crashes.  Any other code you might be tempted to add here
-		 * should probably be in an on_proc_exit callback instead.
+		 * error message, in case things are so hosed that proc_exit crashes.
+		 * Any other code you might be tempted to add here should probably be
+		 * in an on_proc_exit callback instead.
 		 */
 		fflush(stdout);
 		fflush(stderr);
 
 		/*
-		 * If proc_exit is already running, we exit with nonzero exit code
-		 * to indicate that something's pretty wrong.  We also want to
-		 * exit with nonzero exit code if not running under the postmaster
-		 * (for example, if we are being run from the initdb script, we'd
-		 * better return an error status).
+		 * If proc_exit is already running, we exit with nonzero exit code to
+		 * indicate that something's pretty wrong.  We also want to exit with
+		 * nonzero exit code if not running under the postmaster (for example,
+		 * if we are being run from the initdb script, we'd better return an
+		 * error status).
 		 */
 		proc_exit(proc_exit_inprogress || !IsUnderPostmaster);
 	}
@@ -448,8 +440,8 @@ errfinish(int dummy,...)
 	if (elevel >= PANIC)
 	{
 		/*
-		 * Serious crash time. Postmaster will observe nonzero process
-		 * exit status and kill the other backends too.
+		 * Serious crash time. Postmaster will observe nonzero process exit
+		 * status and kill the other backends too.
 		 *
 		 * XXX: what if we are *in* the postmaster?  abort() won't kill our
 		 * children...
@@ -607,7 +599,7 @@ errcode_for_socket_access(void)
 		char		   *fmtbuf; \
 		StringInfoData	buf; \
 		/* Internationalize the error format string */ \
-		fmt = gettext(fmt); \
+		fmt = _(fmt); \
 		/* Expand %m in format string */ \
 		fmtbuf = expand_fmt_string(fmt, edata); \
 		initStringInfo(&buf); \
@@ -984,8 +976,8 @@ CopyErrorData(void)
 	ErrorData  *newedata;
 
 	/*
-	 * we don't increment recursion_depth because out-of-memory here does
-	 * not indicate a problem within the error subsystem.
+	 * we don't increment recursion_depth because out-of-memory here does not
+	 * indicate a problem within the error subsystem.
 	 */
 	CHECK_STACK_DEPTH();
 
@@ -1044,9 +1036,9 @@ void
 FlushErrorState(void)
 {
 	/*
-	 * Reset stack to empty.  The only case where it would be more than
-	 * one deep is if we serviced an error that interrupted construction
-	 * of another message.	We assume control escaped out of that message
+	 * Reset stack to empty.  The only case where it would be more than one
+	 * deep is if we serviced an error that interrupted construction of
+	 * another message.  We assume control escaped out of that message
 	 * construction and won't ever go back.
 	 */
 	errordata_stack_depth = -1;
@@ -1124,7 +1116,7 @@ DebugFileOpen(void)
 					   0666)) < 0)
 			ereport(FATAL,
 					(errcode_for_file_access(),
-			  errmsg("could not open file \"%s\": %m", OutputFileName)));
+				  errmsg("could not open file \"%s\": %m", OutputFileName)));
 		istty = isatty(fd);
 		close(fd);
 
@@ -1138,17 +1130,17 @@ DebugFileOpen(void)
 							OutputFileName)));
 
 		/*
-		 * If the file is a tty and we're running under the postmaster,
-		 * try to send stdout there as well (if it isn't a tty then stderr
-		 * will block out stdout, so we may as well let stdout go wherever
-		 * it was going before).
+		 * If the file is a tty and we're running under the postmaster, try to
+		 * send stdout there as well (if it isn't a tty then stderr will block
+		 * out stdout, so we may as well let stdout go wherever it was going
+		 * before).
 		 */
 		if (istty && IsUnderPostmaster)
 			if (!freopen(OutputFileName, "a", stdout))
 				ereport(FATAL,
 						(errcode_for_file_access(),
-					 errmsg("could not reopen file \"%s\" as stdout: %m",
-							OutputFileName)));
+						 errmsg("could not reopen file \"%s\" as stdout: %m",
+								OutputFileName)));
 	}
 }
 
@@ -1156,6 +1148,39 @@ DebugFileOpen(void)
 
 #ifdef HAVE_SYSLOG
 
+/*
+ * Set or update the parameters for syslog logging
+ */
+void
+set_syslog_parameters(const char *ident, int facility)
+{
+	/*
+	 * guc.c is likely to call us repeatedly with same parameters, so don't
+	 * thrash the syslog connection unnecessarily.	Also, we do not re-open
+	 * the connection until needed, since this routine will get called whether
+	 * or not Log_destination actually mentions syslog.
+	 *
+	 * Note that we make our own copy of the ident string rather than relying
+	 * on guc.c's.  This may be overly paranoid, but it ensures that we cannot
+	 * accidentally free a string that syslog is still using.
+	 */
+	if (syslog_ident == NULL || strcmp(syslog_ident, ident) != 0 ||
+		syslog_facility != facility)
+	{
+		if (openlog_done)
+		{
+			closelog();
+			openlog_done = false;
+		}
+		if (syslog_ident)
+			free(syslog_ident);
+		syslog_ident = strdup(ident);
+		/* if the strdup fails, we will cope in write_syslog() */
+		syslog_facility = facility;
+	}
+}
+
+
 #ifndef PG_SYSLOG_LIMIT
 #define PG_SYSLOG_LIMIT 128
 #endif
@@ -1166,46 +1191,16 @@ DebugFileOpen(void)
 static void
 write_syslog(int level, const char *line)
 {
-	static bool openlog_done = false;
 	static unsigned long seq = 0;
 
 	int			len;
 
+	/* Open syslog connection if not done yet */
 	if (!openlog_done)
 	{
-		int		syslog_fac;
-		char   *syslog_ident;
-
-		if (pg_strcasecmp(Syslog_facility, "LOCAL0") == 0)
-			syslog_fac = LOG_LOCAL0;
-		else if (pg_strcasecmp(Syslog_facility, "LOCAL1") == 0)
-			syslog_fac = LOG_LOCAL1;
-		else if (pg_strcasecmp(Syslog_facility, "LOCAL2") == 0)
-			syslog_fac = LOG_LOCAL2;
-		else if (pg_strcasecmp(Syslog_facility, "LOCAL3") == 0)
-			syslog_fac = LOG_LOCAL3;
-		else if (pg_strcasecmp(Syslog_facility, "LOCAL4") == 0)
-			syslog_fac = LOG_LOCAL4;
-		else if (pg_strcasecmp(Syslog_facility, "LOCAL5") == 0)
-			syslog_fac = LOG_LOCAL5;
-		else if (pg_strcasecmp(Syslog_facility, "LOCAL6") == 0)
-			syslog_fac = LOG_LOCAL6;
-		else if (pg_strcasecmp(Syslog_facility, "LOCAL7") == 0)
-			syslog_fac = LOG_LOCAL7;
-		else
-			syslog_fac = LOG_LOCAL0;
-		/*
-		 * openlog() usually just stores the passed char pointer as-is,
-		 * so we must give it a string that will be unchanged for the life of
-		 * the process.  The Syslog_ident GUC variable does not meet this
-		 * requirement, so strdup() it.  This isn't a memory leak because
-		 * this code is executed at most once per process.
-		 */
-		syslog_ident = strdup(Syslog_ident);
-		if (syslog_ident == NULL)			/* out of memory already!? */
-			syslog_ident = "postgres";
-
-		openlog(syslog_ident, LOG_PID | LOG_NDELAY | LOG_NOWAIT, syslog_fac);
+		openlog(syslog_ident ? syslog_ident : "postgres",
+				LOG_PID | LOG_NDELAY | LOG_NOWAIT,
+				syslog_facility);
 		openlog_done = true;
 	}
 
@@ -1216,13 +1211,12 @@ write_syslog(int level, const char *line)
 	seq++;
 
 	/*
-	 * Our problem here is that many syslog implementations don't handle
-	 * long messages in an acceptable manner. While this function doesn't
-	 * help that fact, it does work around by splitting up messages into
-	 * smaller pieces.
+	 * Our problem here is that many syslog implementations don't handle long
+	 * messages in an acceptable manner. While this function doesn't help that
+	 * fact, it does work around by splitting up messages into smaller pieces.
 	 *
-	 * We divide into multiple syslog() calls if message is too long
-	 * or if the message contains embedded NewLine(s) '\n'.
+	 * We divide into multiple syslog() calls if message is too long or if the
+	 * message contains embedded NewLine(s) '\n'.
 	 */
 	len = strlen(line);
 	if (len > PG_SYSLOG_LIMIT || strchr(line, '\n') != NULL)
@@ -1294,7 +1288,7 @@ write_syslog(int level, const char *line)
 static void
 write_eventlog(int level, const char *line)
 {
-	int eventlevel = EVENTLOG_ERROR_TYPE;
+	int			eventlevel = EVENTLOG_ERROR_TYPE;
 	static HANDLE evtHandle = INVALID_HANDLE_VALUE;
 
 	if (evtHandle == INVALID_HANDLE_VALUE)
@@ -1360,9 +1354,9 @@ log_line_prefix(StringInfo buf)
 	int			i;
 
 	/*
-	 * This is one of the few places where we'd rather not inherit a
-	 * static variable's value from the postmaster.  But since we will,
-	 * reset it when MyProcPid changes.
+	 * This is one of the few places where we'd rather not inherit a static
+	 * variable's value from the postmaster.  But since we will, reset it when
+	 * MyProcPid changes.
 	 */
 	if (log_my_pid != MyProcPid)
 	{
@@ -1387,10 +1381,7 @@ log_line_prefix(StringInfo buf)
 		/* go to char after '%' */
 		i++;
 		if (i >= format_len)
-		{
-			/* format error - ignore it */
-			break;
-		}
+			break;				/* format error - ignore it */
 
 		/* process the option */
 		switch (Log_line_prefix[i])
@@ -1401,7 +1392,7 @@ log_line_prefix(StringInfo buf)
 					const char *username = MyProcPort->user_name;
 
 					if (username == NULL || *username == '\0')
-						username = gettext("[unknown]");
+						username = _("[unknown]");
 					appendStringInfo(buf, "%s", username);
 				}
 				break;
@@ -1411,7 +1402,7 @@ log_line_prefix(StringInfo buf)
 					const char *dbname = MyProcPort->database_name;
 
 					if (dbname == NULL || *dbname == '\0')
-						dbname = gettext("[unknown]");
+						dbname = _("[unknown]");
 					appendStringInfo(buf, "%s", dbname);
 				}
 				break;
@@ -1419,8 +1410,8 @@ log_line_prefix(StringInfo buf)
 				if (MyProcPort)
 				{
 					appendStringInfo(buf, "%lx.%x",
-							   (long) (MyProcPort->session_start.tv_sec),
-							   MyProcPid);
+								   (long) (MyProcPort->session_start.tv_sec),
+									 MyProcPid);
 				}
 				break;
 			case 'p':
@@ -1429,18 +1420,45 @@ log_line_prefix(StringInfo buf)
 			case 'l':
 				appendStringInfo(buf, "%ld", log_line_number);
 				break;
-			case 't':
+			case 'm':
 				{
 					/*
-					 * Note: for %t and %s we deliberately use the C
-					 * library's strftime/localtime, and not the
-					 * equivalent functions from src/timezone.	This
-					 * ensures that all backends will report log entries
-					 * in the same timezone, namely whatever C-library
-					 * setting they inherit from the postmaster.  If we
-					 * used src/timezone then local settings of the
-					 * TimeZone GUC variable would confuse the log.
+					 * Note: for %m, %t, and %s we deliberately use the C
+					 * library's strftime/localtime, and not the equivalent
+					 * functions from src/timezone.  This ensures that all
+					 * backends will report log entries in the same timezone,
+					 * namely whatever C-library setting they inherit from the
+					 * postmaster.	If we used src/timezone then local
+					 * settings of the TimeZone GUC variable would confuse the
+					 * log.
 					 */
+					time_t		stamp_time;
+					char		strfbuf[128],
+								msbuf[8];
+					struct timeval tv;
+
+					gettimeofday(&tv, NULL);
+					stamp_time = tv.tv_sec;
+
+					strftime(strfbuf, sizeof(strfbuf),
+					/* leave room for milliseconds... */
+					/* Win32 timezone names are too long so don't print them. */
+#ifndef WIN32
+							 "%Y-%m-%d %H:%M:%S     %Z",
+#else
+							 "%Y-%m-%d %H:%M:%S     ",
+#endif
+							 localtime(&stamp_time));
+
+					/* 'paste' milliseconds into place... */
+					sprintf(msbuf, ".%03d", (int) (tv.tv_usec / 1000));
+					strncpy(strfbuf + 19, msbuf, 4);
+
+					appendStringInfoString(buf, strfbuf);
+				}
+				break;
+			case 't':
+				{
 					time_t		stamp_time = time(NULL);
 					char		strfbuf[128];
 
@@ -1471,7 +1489,7 @@ log_line_prefix(StringInfo buf)
 				if (MyProcPort)
 				{
 					const char *psdisp;
-					int		displen;
+					int			displen;
 
 					psdisp = get_ps_display(&displen);
 					appendStringInfo(buf, "%.*s", displen, psdisp);
@@ -1487,6 +1505,10 @@ log_line_prefix(StringInfo buf)
 										 MyProcPort->remote_port);
 				}
 				break;
+			case 'h':
+				if (MyProcPort && MyProcPort->remote_host)
+					appendStringInfo(buf, "%s", MyProcPort->remote_host);
+				break;
 			case 'q':
 				/* in postmaster and friends, stop if %q is seen */
 				/* in a backend, just ignore */
@@ -1512,6 +1534,26 @@ log_line_prefix(StringInfo buf)
 	}
 }
 
+/*
+ * Unpack MAKE_SQLSTATE code. Note that this returns a pointer to a
+ * static buffer.
+ */
+char *
+unpack_sql_state(int sql_state)
+{
+	static char buf[12];
+	int			i;
+
+	for (i = 0; i < 5; i++)
+	{
+		buf[i] = PGUNSIXBIT(sql_state);
+		sql_state >>= 6;
+	}
+
+	buf[i] = '\0';
+	return buf;
+}
+
 
 /*
  * Write error report to server's log
@@ -1527,32 +1569,18 @@ send_message_to_server_log(ErrorData *edata)
 	appendStringInfo(&buf, "%s:  ", error_severity(edata->elevel));
 
 	if (Log_error_verbosity >= PGERROR_VERBOSE)
-	{
-		/* unpack MAKE_SQLSTATE code */
-		char		tbuf[12];
-		int			ssval;
-		int			i;
-
-		ssval = edata->sqlerrcode;
-		for (i = 0; i < 5; i++)
-		{
-			tbuf[i] = PGUNSIXBIT(ssval);
-			ssval >>= 6;
-		}
-		tbuf[i] = '\0';
-		appendStringInfo(&buf, "%s: ", tbuf);
-	}
+		appendStringInfo(&buf, "%s: ", unpack_sql_state(edata->sqlerrcode));
 
 	if (edata->message)
 		append_with_tabs(&buf, edata->message);
 	else
-		append_with_tabs(&buf, gettext("missing error text"));
+		append_with_tabs(&buf, _("missing error text"));
 
 	if (edata->cursorpos > 0)
-		appendStringInfo(&buf, gettext(" at character %d"),
+		appendStringInfo(&buf, _(" at character %d"),
 						 edata->cursorpos);
 	else if (edata->internalpos > 0)
-		appendStringInfo(&buf, gettext(" at character %d"),
+		appendStringInfo(&buf, _(" at character %d"),
 						 edata->internalpos);
 
 	appendStringInfoChar(&buf, '\n');
@@ -1562,28 +1590,28 @@ send_message_to_server_log(ErrorData *edata)
 		if (edata->detail)
 		{
 			log_line_prefix(&buf);
-			appendStringInfoString(&buf, gettext("DETAIL:  "));
+			appendStringInfoString(&buf, _("DETAIL:  "));
 			append_with_tabs(&buf, edata->detail);
 			appendStringInfoChar(&buf, '\n');
 		}
 		if (edata->hint)
 		{
 			log_line_prefix(&buf);
-			appendStringInfoString(&buf, gettext("HINT:  "));
+			appendStringInfoString(&buf, _("HINT:  "));
 			append_with_tabs(&buf, edata->hint);
 			appendStringInfoChar(&buf, '\n');
 		}
 		if (edata->internalquery)
 		{
 			log_line_prefix(&buf);
-			appendStringInfoString(&buf, gettext("QUERY:  "));
+			appendStringInfoString(&buf, _("QUERY:  "));
 			append_with_tabs(&buf, edata->internalquery);
 			appendStringInfoChar(&buf, '\n');
 		}
 		if (edata->context)
 		{
 			log_line_prefix(&buf);
-			appendStringInfoString(&buf, gettext("CONTEXT:  "));
+			appendStringInfoString(&buf, _("CONTEXT:  "));
 			append_with_tabs(&buf, edata->context);
 			appendStringInfoChar(&buf, '\n');
 		}
@@ -1593,27 +1621,26 @@ send_message_to_server_log(ErrorData *edata)
 			if (edata->funcname && edata->filename)
 			{
 				log_line_prefix(&buf);
-				appendStringInfo(&buf, gettext("LOCATION:  %s, %s:%d\n"),
+				appendStringInfo(&buf, _("LOCATION:  %s, %s:%d\n"),
 								 edata->funcname, edata->filename,
 								 edata->lineno);
 			}
 			else if (edata->filename)
 			{
 				log_line_prefix(&buf);
-				appendStringInfo(&buf, gettext("LOCATION:  %s:%d\n"),
+				appendStringInfo(&buf, _("LOCATION:  %s:%d\n"),
 								 edata->filename, edata->lineno);
 			}
 		}
 	}
 
 	/*
-	 * If the user wants the query that generated this error logged, do
-	 * it.
+	 * If the user wants the query that generated this error logged, do it.
 	 */
 	if (edata->elevel >= log_min_error_statement && debug_query_string != NULL)
 	{
 		log_line_prefix(&buf);
-		appendStringInfoString(&buf, gettext("STATEMENT:  "));
+		appendStringInfoString(&buf, _("STATEMENT:  "));
 		append_with_tabs(&buf, debug_query_string);
 		appendStringInfoChar(&buf, '\n');
 	}
@@ -1667,24 +1694,23 @@ send_message_to_server_log(ErrorData *edata)
 #endif   /* WIN32 */
 
 	/* Write to stderr, if enabled */
-	if ((Log_destination & LOG_DESTINATION_STDERR) || whereToSendOutput == Debug)
+	if ((Log_destination & LOG_DESTINATION_STDERR) || whereToSendOutput == DestDebug)
 	{
 #ifdef WIN32
+
 		/*
 		 * In a win32 service environment, there is no usable stderr. Capture
 		 * anything going there and write it to the eventlog instead.
 		 *
-		 * If stderr redirection is active, it's ok to write to stderr
-		 * because that's really a pipe to the syslogger process.
+		 * If stderr redirection is active, it's ok to write to stderr because
+		 * that's really a pipe to the syslogger process. Unless we're in the
+		 * postmaster, and the syslogger process isn't started yet.
 		 */
-		if (pgwin32_is_service() && (!redirection_done || am_syslogger) )
+		if ((!Redirect_stderr || am_syslogger || (!IsUnderPostmaster && SysLoggerPID==0)) && pgwin32_is_service())
 			write_eventlog(edata->elevel, buf.data);
 		else
 #endif
-			if (redirection_done && !am_syslogger)
-				write_pipe_chunks(fileno(stderr), buf.data, buf.len);
-			else
-				write(fileno(stderr), buf.data, buf.len);
+			fprintf(stderr, "%s", buf.data);
 	}
 
 	/* If in the syslogger process, try to write messages direct to file */
@@ -1694,37 +1720,6 @@ send_message_to_server_log(ErrorData *edata)
 	pfree(buf.data);
 }
 
-/*
- * Send data to the syslogger using the chunked protocol
- */
-static void
-write_pipe_chunks(int fd, char *data, int len)
-{
-	PipeProtoChunk p;
-
-	Assert(len > 0);
-
-	p.proto.nuls[0] = p.proto.nuls[1] = '\0';
-	p.proto.pid = MyProcPid;
-
-	/* write all but the last chunk */
-	while (len > PIPE_MAX_PAYLOAD)
-	{
-		p.proto.is_last = 'f';
-		p.proto.len = PIPE_MAX_PAYLOAD;
-		memcpy(p.proto.data, data, PIPE_MAX_PAYLOAD);
-		write(fd, &p, PIPE_HEADER_SIZE + PIPE_MAX_PAYLOAD);
-		data += PIPE_MAX_PAYLOAD;
-		len -= PIPE_MAX_PAYLOAD;
-	}
-
-	/* write the last chunk */
-	p.proto.is_last = 't';
-	p.proto.len = len;
-	memcpy(p.proto.data, data, len);
-	write(fd, &p, PIPE_HEADER_SIZE + len);
-}
-
 
 /*
  * Write error report to client
@@ -1764,7 +1759,7 @@ send_message_to_frontend(ErrorData *edata)
 		if (edata->message)
 			pq_sendstring(&msgbuf, edata->message);
 		else
-			pq_sendstring(&msgbuf, gettext("missing error text"));
+			pq_sendstring(&msgbuf, _("missing error text"));
 
 		if (edata->detail)
 		{
@@ -1840,13 +1835,13 @@ send_message_to_frontend(ErrorData *edata)
 		if (edata->message)
 			appendStringInfoString(&buf, edata->message);
 		else
-			appendStringInfoString(&buf, gettext("missing error text"));
+			appendStringInfoString(&buf, _("missing error text"));
 
 		if (edata->cursorpos > 0)
-			appendStringInfo(&buf, gettext(" at character %d"),
+			appendStringInfo(&buf, _(" at character %d"),
 							 edata->cursorpos);
 		else if (edata->internalpos > 0)
-			appendStringInfo(&buf, gettext(" at character %d"),
+			appendStringInfo(&buf, _(" at character %d"),
 							 edata->internalpos);
 
 		appendStringInfoChar(&buf, '\n');
@@ -1859,12 +1854,12 @@ send_message_to_frontend(ErrorData *edata)
 	pq_endmessage(&msgbuf);
 
 	/*
-	 * This flush is normally not necessary, since postgres.c will flush
-	 * out waiting data when control returns to the main loop. But it
-	 * seems best to leave it here, so that the client has some clue what
-	 * happened if the backend dies before getting back to the main loop
-	 * ... error/notice messages should not be a performance-critical path
-	 * anyway, so an extra flush won't hurt much ...
+	 * This flush is normally not necessary, since postgres.c will flush out
+	 * waiting data when control returns to the main loop. But it seems best
+	 * to leave it here, so that the client has some clue what happened if the
+	 * backend dies before getting back to the main loop ... error/notice
+	 * messages should not be a performance-critical path anyway, so an extra
+	 * flush won't hurt much ...
 	 */
 	pq_flush();
 }
@@ -1899,9 +1894,9 @@ expand_fmt_string(const char *fmt, ErrorData *edata)
 			if (*cp == 'm')
 			{
 				/*
-				 * Replace %m by system error string.  If there are any
-				 * %'s in the string, we'd better double them so that
-				 * vsnprintf won't misinterpret.
+				 * Replace %m by system error string.  If there are any %'s in
+				 * the string, we'd better double them so that vsnprintf won't
+				 * misinterpret.
 				 */
 				const char *cp2;
 
@@ -1946,8 +1941,8 @@ useful_strerror(int errnum)
 	str = strerror(errnum);
 
 	/*
-	 * Some strerror()s return an empty string for out-of-range errno.
-	 * This is ANSI C spec compliant, but not exactly useful.
+	 * Some strerror()s return an empty string for out-of-range errno. This is
+	 * ANSI C spec compliant, but not exactly useful.
 	 */
 	if (str == NULL || *str == '\0')
 	{
@@ -1956,7 +1951,7 @@ useful_strerror(int errnum)
 		 * expanded.
 		 */
 		snprintf(errorstr_buf, sizeof(errorstr_buf),
-				 gettext("operating system error %d"), errnum);
+				 _("operating system error %d"), errnum);
 		str = errorstr_buf;
 	}
 
@@ -1979,29 +1974,29 @@ error_severity(int elevel)
 		case DEBUG3:
 		case DEBUG4:
 		case DEBUG5:
-			prefix = gettext("DEBUG");
+			prefix = _("DEBUG");
 			break;
 		case LOG:
 		case COMMERROR:
-			prefix = gettext("LOG");
+			prefix = _("LOG");
 			break;
 		case INFO:
-			prefix = gettext("INFO");
+			prefix = _("INFO");
 			break;
 		case NOTICE:
-			prefix = gettext("NOTICE");
+			prefix = _("NOTICE");
 			break;
 		case WARNING:
-			prefix = gettext("WARNING");
+			prefix = _("WARNING");
 			break;
 		case ERROR:
-			prefix = gettext("ERROR");
+			prefix = _("ERROR");
 			break;
 		case FATAL:
-			prefix = gettext("FATAL");
+			prefix = _("FATAL");
 			break;
 		case PANIC:
-			prefix = gettext("PANIC");
+			prefix = _("PANIC");
 			break;
 		default:
 			prefix = "???";
@@ -2042,13 +2037,12 @@ write_stderr(const char *fmt,...)
 {
 	va_list		ap;
 
-	fmt = gettext(fmt);
+	fmt = _(fmt);
 
 	va_start(ap, fmt);
 #ifndef WIN32
 	/* On Unix, we just fprintf to stderr */
 	vfprintf(stderr, fmt, ap);
-	fflush(stderr);
 #else
 
 	/*
@@ -2064,11 +2058,8 @@ write_stderr(const char *fmt,...)
 		write_eventlog(EVENTLOG_ERROR_TYPE, errbuf);
 	}
 	else
-	{
 		/* Not running as service, write to stderr */
 		vfprintf(stderr, fmt, ap);
-		fflush(stderr);
-	}
 #endif
 	va_end(ap);
 }
